@@ -45,7 +45,7 @@ enum {
 	LAST_SIGNAL
 };
 
-static gint color_combo_signals [LAST_SIGNAL] = { 0, };
+static guint color_combo_signals [LAST_SIGNAL] = { 0, };
 
 static GtkObjectClass *color_combo_parent_class;
 
